#include <u.h>
#include <libc.h>
#include <auth.h>
#include <mp.h>
#include <libsec.h>
#include <der.h>
#include <fcall.h>
#include <thread.h>
#include <9p.h>
#include "ldap.h"

struct Cache {
	Cache *next;
	int refs;
	char *dn;
	int num;
	Dir *dirs;
};


static Dir 
copyone(Dir *old)
{
	Dir dir;

	dir = *old;
	if(dir.name)		dir.name = estrdup9p(dir.name);
	if(dir.muid)		dir.muid = estrdup9p(dir.muid);
	if(dir.uid)			dir.uid = estrdup9p(dir.uid);
	if(dir.gid)			dir.gid = estrdup9p(dir.gid);
	return dir;
}

static Dir *
copyall(Dir *old, int num)
{
	Dir *dirs, *d;

	dirs = emalloc9p(sizeof(Dir) * num);
	for(d = dirs; d < &dirs[num]; d++){
		*d = *old++;
		if(d->name)		d->name = estrdup9p(d->name);
		if(d->muid)		d->muid = estrdup9p(d->muid);
		if(d->uid)		d->uid = estrdup9p(d->uid);
		if(d->gid)		d->gid = estrdup9p(d->gid);
	}
	return dirs;
}

static void
freeall(Dir *dirs, int num)
{
	Dir *d;

	for(d = dirs; d < &dirs[num]; d++){
		free(d->name);
		free(d->muid);
		free(d->uid);
		free(d->gid);
	}
	free(dirs);
}

static int
eq_tail(char *full, char *tail)
{
	int nf, nt;

	nf = strlen(full);
	nt = strlen(tail);
	return strncmp(full + (nf-nt), tail, nt) == 0;
}

int
cached_stat(Cache *cache, char *dn, Dir *d)
{
	int i;
	Cache *c, *hit;
	char *tail, *head;

	if((tail = strchr(dn, ',')) != nil)
		tail++;
	else
		tail = strchr(dn, 0);

	if((head = strchr(dn, '=')) == nil)
		sysfatal("cached_stat: '%s' DN does not contain a '='", dn);
	head++;

	hit = nil;
	for(c = cache; c; c = c->next){
		if(strcmp(tail, c->dn) == 0){		/* exact match found */
			hit = c;
			break;
		}
		if(strcmp("", c->dn) == 0)			/* by default, guess at the root dir */
			hit = c;
	}
	if(hit == nil)							/* cannot find it */
		return -1;

	for(i = 0; i < hit->num; i++)
		if(strncmp(head, hit->dirs[i].name, strlen(hit->dirs[i].name)) == 0){
			*d = copyone(&hit->dirs[i]);
			return 0;
		}
	return -1;
}

int
cached_readdir(Cache *cache, char *dn, Dir **dirs)
{
	Cache *c;

	for(c = cache; c; c = c->next)
		if(strcmp(c->dn, dn) == 0){
			*dirs = copyall(c->dirs, c->num);
			return c->num;
		}
	return -1;
}

char *
cached_walk(Cache *cache, char *dn, char *name, Qid *q)
{
	int i;
	Cache *c;

	for(c = cache; c; c = c->next)
		if(strcmp(c->dn, dn) == 0)
			for(i = 0; i < c->num; i++)
				if(strcmp(c->dirs[i].name, name) == 0){
					*q = c->dirs[i].qid;
					return estrdup9p(c->dirs[i].muid);
				}
	return nil;
}

Cache *
cache_fill(Cache *cache, char *dn, int num, Dir *dirs)
{
	Cache *c;

	for(c = cache; c; c = c->next)
		if(strcmp(c->dn, dn) == 0){
			fprint(2, "cache_fill dn=%s already known, num: %d->%d\n", dn, c->num, num);
			return cache;
		}

	c = emalloc9p(sizeof(Cache));
	c->dn = estrdup9p(dn);
	c->num = num;
	c->refs = 1;
	c->dirs = copyall(dirs, num);

	c->next = cache;
	return c;
}

int
cache_decref(Cache *cache, char *dn)
{
	Cache *c;

	for(c = cache; c; c = c->next)
		if(strcmp(c->dn, dn) == 0){
			if(c->refs > 0)
				c->refs--;
			return 0;
		}
	return -1;
}

int
cache_incref(Cache *cache, char *dn)
{
	Cache *c;

	for(c = cache; c; c = c->next)
		if(strcmp(c->dn, dn) == 0){
			c->refs++;
			return 0;
		}
	return -1;
}

