/*
 * decode dnsRecord ldap attributes
 * see https://www.indented.co.uk/mapping-the-dnsrecord-attribute
 */

#include <u.h>
#include <libc.h>
#include <auth.h>
#include <mp.h>
#include <libsec.h>
#include <der.h>
#include <fcall.h>
#include <thread.h>
#include <9p.h>
#include <ctype.h>
#include <ip.h>
#include "ldap.h"

enum {						 	/* values for rank field */
	CacheBit = 1,				/* from the cache */
	RootHint = 8,				/* preconfigured root hint */
	OutsideGlue = 32,			/* not used */
	CacheNAAdditional = 49,		/* from the additional section of a nonauthoritative response */
	CacheNAAuthority = 65,		/* from the authority section of a nonauthoritative response */
	CacheAAdditional = 81,		/* from the additional section of an authoritative response */
	CacheNAAnswer = 97,			/* from the answer section of a nonauthoritative response */
	CacheAAuthority = 113,		/* from the authority section of an authoritative response */
	Glue = 128,					/* glue record in an authoritative zone */
	NSGlue = 130,				/* delegation (type NS) record in an authoritative zone */
	CacheAAnswer = 193,			/* from the answer section of an authoritative response */
	Zone = 240					/* authoritative zone */
};

enum {
	Fqdnmax		= 255,

	Cin			= 1,		/* internet; other classes are effectively unused */

	Ta			= 1,		/* a host address */
	Tns			= 2,		/* an authoritative name server */
	Tmd			= 3,		/* a mail destination (Obsolete - use MX) */
	Tmf			= 4,		/* a mail forwarder (Obsolete - use MX) */
	Tcname		= 5,		/* the canonical name for an alias */
	Tsoa		= 6,		/* marks the start of a zone of authority */
	Tmb			= 7,		/* a mailbox domain name (EXPERIMENTAL) */
	Tmg			= 8,		/* a mail group member (EXPERIMENTAL) */
	Tmr			= 9,		/* a mail rename domain name (EXPERIMENTAL) */
	Tnull		= 10,		/* a null RR (EXPERIMENTAL) */
	Twks		= 11,		/* a well known service description */
	Tptr		= 12,		/* a domain name pointer */
	Thinfo		= 13,		/* host information */
	Tminfo		= 14,		/* mailbox or mail list information */
	Tmx			= 15,		/* mail exchange */
	Ttxt		= 16,		/* text strings */
	Trp			= 17,		/* for Responsible Person */
	Tafsdb		= 18,		/* for AFS Data Base location */
	Tx25		= 19,		/* for X.25 PSDN address */
	Tisdn		= 20,		/* for ISDN address */
	Trt			= 21,		/* for Route Through */
	Tnsap		= 22,		/* for NSAP address, NSAP style A record */
	Tnsap_ptr	= 23,		/* for domain name pointer, NSAP style */
	Tsig		= 24,		/* for security signature */
	Tkey		= 25,		/* for security key */
	Tpx			= 26,		/* X.400 mail mapping information */
	Tgpos		= 27,		/* Geographical Position */
	Taaaa		= 28,		/* IP6 Address */
	Tloc		= 29,		/* Location Information */
	Tnxt		= 30,		/* Next Domain (obsolete) */
	Teid		= 31,		/* Endpoint Identifier */
	Tnimloc		= 32,		/* Nimrod Locator */
	Tsrv		= 33,		/* Server Selection */
	Tatma		= 34,		/* ATM Address */
	Tnaptr		= 35,		/* Naming Authority Pointer */
	Tkx			= 36,		/* Key Exchanger */
	Tcert		= 37,
	Ta6			= 38,
	Tdname		= 39,
	Tsink		= 40,
	Topt		= 41,
	Tapl		= 42,
	Tds			= 43,		/* Delegation Signer */
	Tsshfp		= 44,		/* SSH Key Fingerprint */
	Trrsig		= 46,
	Tnsec		= 47,
	Tdnskey		= 48,
	Tuinfo		= 100,
	Tuid		= 101,
	Tgid		= 102,
	Tunspec		= 103,
	Taddrs		= 248,
	Ttkey		= 249,
	Ttsig		= 250,		/* Transaction Signature */
	Tixfr		= 251,		/* Incremental transfer */
	Taxfr		= 252,		/* transfer of an entire zone */
	Tmailb		= 253,		/* mailbox-related RRs (MB, MG or MR) */
	Tmaila		= 254,		/* mail agent RRs (Obsolete - see MX) */
	Tall		= 255,		/* A request for all records */
	Twins		= 0xff01,	/* MS WINS server */
	Twinsr		= 0xff02,	/* MS wins reverse lookup */
};

static char *TypeName[] = {
	[Ta]		"a",
	[Tns]		"ns",
	[Tmd]		"md",
	[Tmf]		"mf",
	[Tcname]	"cname",
	[Tsoa]		"soa",
	[Tmb]		"mb",
	[Tmg]		"mg",
	[Tmr]		"mr",
	[Tnull]		"null",
	[Twks]		"wks",
	[Tptr]		"ptr",
	[Thinfo]	"hinfo",
	[Tminfo]	"minfo",
	[Tmx]		"mx",
	[Ttxt]		"txt",
	[Trp]		"rp",
	[Tafsdb]	"afsdb",
	[Tx25]		"x25",
	[Tisdn]		"isdn",
	[Trt]		"rt",
	[Tnsap]		"nsap",
	[Tnsap_ptr]	"nsap_ptr",
	[Tsig]		"sig",
	[Tkey]		"key",
	[Tpx]		"px",
	[Tgpos]		"gpos",
	[Taaaa]		"aaaa",
	[Tloc]		"loc",
	[Tnxt]		"nxt",
	[Teid]		"eid",
	[Tnimloc]	"nimloc",
	[Tsrv]		"srv",
	[Tatma]		"atma",
	[Tnaptr]	"naptr",
	[Tkx]		"kx",
	[Tcert]		"cert",
	[Ta6]		"a6",
	[Tdname]	"dname",
	[Tsink]		"sink",
	[Topt]		"opt",
	[Tapl]		"apl",
	[Tds]		"ds",
	[Tsshfp]	"sshfp",
	[Trrsig]	"rrsig",
	[Tnsec]		"nsec",
	[Tdnskey]	"dnskey",
	[Tuinfo]	"uinfo",
	[Tuid]		"uid",
	[Tgid]		"gid",
	[Tunspec]	"unspec",
	[Taddrs]	"addrs",
	[Ttkey]		"tkey",
	[Ttsig]		"tsig",
	[Tixfr]		"ixfr",
	[Taxfr]		"axfr",
	[Tmailb]	"mailb",
	[Tmaila]	"maila",
	[Tall]		"all",
};


static int
g8(uchar **p)
{
	return *(*p)++;
}

static ushort
g16(uchar **p)
{
	*p += 2;
	return (*p)[-1] << 8 | (*p)[-2];
}

static int
g32(uchar **p)
{
	*p += 4;
	return (*p)[-1] << 24 | (*p)[-2] << 16 | (*p)[-3] << 8 | (*p)[-4];
}

void
gname(uchar **pp, uchar *buf, char *s)
{
	char *last = s;
	uchar *q;
	int c, total, chars, w, words;

	total = g8(pp);
	words = g8(pp);
	for(w = 0; w < words; w++){
		chars = g8(pp);
		if(chars == 0)			/* empty word is the end */
			break;

		if(chars & 0xc0){
			(*pp)--;
			chars = g16(pp);
			q = buf + (chars & 0x3fff) + 2;  /* +2 to skip packet len */
			gname(&q, buf, s);
			return;
		}

		for(c = 0; c < chars; c++)
			*s++ = *(*pp)++;

		last = s;
		*s++ = '.';
	}
	*last = 0;

	USED(total);
}

static void
skip(uchar **p, int len)
{
	if(Debug)
		fprint(2, "skipped %d bytes at end of record", len);
	*p += len;
}

static void *
gmem(uchar **p, int len)
{
	char *s;

	assert(s = malloc(len + 1));
	memmove(s, *p, len);
	*p += len;
	s[len] = '\0';		/* unnecessary courtesy */
	return s;
}

static int
xd(Fmt *f, uchar *buf, int len)
{
	int i, first, rc;

	rc = 0;
	first = 1;
	for(i = 0; i < len; i++){
		if(! first)
			fmtprint(f, " ");
		rc += fmtprint(f, "%02x", buf[i]);
		first = 0;
	}
	return rc;
}

static int
pr(Fmt *f, uchar **pp, uchar *buf, int vlen, int type)
{
	int rc, n;
	uchar *ad = nil;
	char *p = nil, s[Fqdnmax +1];

	rc = 0;
	switch(type){
	case Ta:
		ad = gmem(pp, vlen);
		rc += fmtprint(f, "ip=%V ", ad);
		break;
	case Taaaa:
		ad = gmem(pp, vlen);
		rc += fmtprint(f, "ip=%I ", ad);
		break;
	case Thinfo:
		p = gmem(pp, vlen);
		rc += fmtprint(f, "hinfo='%s' ", p);
		break;
	case Ttxt:
		do{
			n = g8(pp);
			p = gmem(pp, n);
			rc += fmtprint(f, "txtrr='%.*s' ", n, p);
			free(p);
			p = nil;
			vlen -= n+1;
		}while(vlen > 0);
		break;
	case Trp:
		gname(pp, buf, s);
		if((p = strchr(s, '.')) != nil)
			*p = '@';
		p = nil;		/* don't free p */
		rc += fmtprint(f, "contact=%s ", s);
		break;
	case Tns:
		gname(pp, buf, s);
		rc += fmtprint(f, "ns=%s ", s);
		break;
	case Tptr:
		gname(pp, buf, s);
		rc += fmtprint(f, "ptr=%s ", s);
		break;
	case Tcname:
		gname(pp, buf, s);
		rc += fmtprint(f, "cname=%s ", s);
		break;
	case Tmx:
		n = g16(pp);
		gname(pp, buf, s);
		rc += fmtprint(f, "mx=%s pref=%ud ", s, n);
		break;
	case Tsoa:
	 	rc += fmtprint(f, "serial=%ud ", g32(pp));
	 	rc += fmtprint(f, "refresh=%ud ", g32(pp));
	 	rc += fmtprint(f, "retry=%ud ", g32(pp));
	 	rc += fmtprint(f, "expire=%ud ", g32(pp));
	 	rc += fmtprint(f, "min=%ud ", g32(pp));
		gname(pp, buf, s);
		rc += fmtprint(f, "ns=%s ", s);

		gname(pp, buf, s);
		if((p = strchr(s, '.')) != nil)
			*p = '@';
		p = nil;		/* don't free p */
		rc += fmtprint(f, "mbox=%s ", s);
		break;
	case Tsrv:
	 	rc += fmtprint(f, "priority=%ud ", g16(pp));
	 	rc += fmtprint(f, "weight=%ud ", g16(pp));
		rc += fmtprint(f, "port=%ud ", g16(pp));
		gname(pp, buf, s);
		rc += fmtprint(f, "dom=%s ", s);
		break;
	case Tgpos:
		rc += fmtprint(f, "longitude=%d ", g16(pp));
		rc += fmtprint(f, "latitude=%d ", g16(pp));
		rc += fmtprint(f, "altitude=%d", g16(pp));
		break;
	case Tx25:
		p = gmem(pp, vlen);
		rc += fmtprint(f, "x25=%s ", p);
		break;
	case Tisdn:
		p = gmem(pp, vlen);
		rc += fmtprint(f, "isdn=%s ", p);
		break;
	case Twins:
		rc += fmtprint(f, "replication=%s ", g32(pp)? "none": "active");
		rc += fmtprint(f, "timeout=%d ", g32(pp));
		rc += fmtprint(f, "cache=%d ", g32(pp));
		n = g32(pp);
		while(n--){
			ad = gmem(pp, 4);
			rc += fmtprint(f, "server=%V ", ad);
			free(ad);
			ad = nil;
		}
		rc += fmtprint(f, "");
		break;
	case Twinsr:
		rc += fmtprint(f, "replication=%s ", g32(pp)? "none": "active");
		rc += fmtprint(f, "timeout=%d ", g32(pp));
		rc += fmtprint(f, "cache=%d ", g32(pp));
		gname(pp, buf, s);
		rc += fmtprint(f, "domain=%s ", s);
		break;
	default:
		if(type > 0 && type < nelem(TypeName) && TypeName[type])
			rc += fmtprint(f, "rr=%s unsupported: ", TypeName[type]);
		else{
			rc += fmtprint(f, "rr=%d unknown: ", type);
		}
		rc += xd(f, *pp, vlen);
		skip(pp, vlen);
		break;
	}
	free(p);
	free(ad);
	return rc;
}

int
dnsRecord(Fmt *f, Bytes *b, int, void *)
{
	uchar *buf;
	int i, len, vlen, type, vers, rank, flags, serno, ttl, zero, time;

	buf = b->data;
	len = b->len;

	vlen = g16(&buf);		/* length of record */
	type = g16(&buf);		/* record type */
	vers = g8(&buf);		/* version, always 5 */
	rank = g8(&buf);		/* rank, see enum above */
	flags = g16(&buf);		/* flags, always zero */
	serno = g32(&buf);		/* serial number */
	ttl = g32(&buf);		/* time to live */
	zero = g32(&buf);		/* unused */
	time = g32(&buf);		/* timestamp */

	len -= buf - b->data;

	if(Debug && strstr(Debug, "dns")){
		fprint(2, "dns: len=%d vlen=%d type=%d vers=%d rank=%x flags=%x serno=%d ttl=%d zero=%d time=%d\n",
			len, vlen, type, vers, rank, flags, serno, ttl, zero, time);

		fprint(2, "dns: ");
		for(i = 0; i < len && i < vlen; i++)
			fprint(2, " %02x", buf[i]);
		fprint(2, "\n");
	}

	return pr(f, &buf, b->data, len, type);

}

