#include <u.h>
#include <libc.h>
#include <auth.h>
#include <mp.h>
#include <fcall.h>
#include <thread.h>
#include <libsec.h>
#include <pool.h>
#include <9p.h>
#include <der.h>
#include <ctype.h>
#include "ldap.h"

typedef struct Num2Str Num2Str;
typedef struct Str2Str Str2Str;

struct Num2Str {
	uvlong n;
	char *s;
};
struct Str2Str {
	char *x;
	char *y;
};

// see here:  http://www.selfadsi.org/deep-inside/ad-security-descriptors.htm

static Num2Str AccountControl[] = {
	{ 0x02,			"account disabled" },
	{ 0x08,			"home directory requred" },
	{ 0x10,			"account locked out" },
	{ 0x20,			"passwd not needed" },
	{ 0x40,			"passwd cannot be changed" },
	{ 0x80,			"encrypted text passwds allowed" },
	{ 0x200,		"normal account" },
	{ 0x800,		"interdomain trust account" },
	{ 0x1000,		"workstation trust account" },
	{ 0x2000,		"server trust account" },
	{ 0x10000,		"password never expires" },
	{ 0x20000,		"majority node set account" },
	{ 0x40000,		"smartcard required" },
	{ 0x80000,		"trusted for delegation" },
	{ 0x100000,		"not delegated" },
	{ 0x200000,		"use DES key only" },
	{ 0x400000,		"does not require preauth" },
	{ 0x800000,		"password expired" },
	{ 0x1000000,	"trusted to auth for delegation" },
	{ 0x2000000,	"no auth required" },
	{ 0x4000000,	"partial secrets account" },
};

static Num2Str ExchangeVersion[] = {
	{ 0,					"Exchange 2003" },
	{ 4535486012416LL,		"Exchange 2007, v8" },
	{ 44220983382016LL,		"Exchange 2010, v14" },
	{ 88218628259840LL,		"Exchange 2013, v15" },
	{ 1125899906842624LL,	"Exchange 2016, v15.01" },
};

static Num2Str AccountType[] = {
	{ 0x00000000,	"domain object" },
	{ 0x10000000,	"group object" },
	{ 0x10000001,	"non security group object" },
	{ 0x20000000,	"alias object" },
	{ 0x20000001,	"non security alias object" },
	{ 0x30000000,	"user object" },
	{ 0x30000000,	"normal user account" },
	{ 0x30000001,	"machine account" },
	{ 0x30000002,	"trust account" },
	{ 0x40000000,	"app basic group" },
	{ 0x40000001,	"app query group" },
};

static Num2Str RecipentType[] = {
	{ -2147483642,	"MailUser (RemoteUserMailbox)" },
	{ -2147481850,	"MailUser (RemoteRoomMailbox)" },
	{ -2147481594,	"MailUser (RemoteEquipmentMailbox)" },
	{ 0,			"UserMailbox (shared)" },
	{ 1,			"MailUniversalDistributionGroup" },
	{ 6,			"MailContact" },
	{ 7,			"UserMailbox (room)" },
	{ 8,			"UserMailbox (equipment)" },
	{ 1073741824,	"UserMailbox" },
	{ 1073741833,	"MailUniversalSecurityGroup" },
};

static Str2Str WellKnownSids[] = {
	{ "S-1-0",					"Null Authority" },
	{ "S-1-0-0",				"Nobody" },
	{ "S-1-1",					"World Authority" },
	{ "S-1-1-0",				"Everyone" },
	{ "S-1-2",					"Local Authority" },
	{ "S-1-2-0",				"Local" },
	{ "S-1-2-1",				"Console Logon" },
	{ "S-1-3",					"Creator Authority" },
	{ "S-1-3-0",				"Creator Owner" },
	{ "S-1-3-1",				"Creator Group" },
	{ "S-1-3-2",				"Creator Owner Server" },
	{ "S-1-3-3",				"Creator Group Server" },
	{ "S-1-3-4",				"Owner Rights" },
	{ "S-1-4",					"Non-unique Authority" },
	{ "S-1-5",					"NT Authority" },
	{ "S-1-5-1",				"Dialup" },
	{ "S-1-5-2",				"Network" },
	{ "S-1-5-3",				"Batch" },
	{ "S-1-5-4",				"Interactive" },
	{ "S-1-5-6",				"Service" },
	{ "S-1-5-7",				"Anonymous" },
	{ "S-1-5-8",				"Proxy" },
	{ "S-1-5-9",				"Enterprise Domain Controllers" },
	{ "S-1-5-10",				"Principal Self" },
	{ "S-1-5-11",				"Authenticated Users" },
	{ "S-1-5-12",				"Restricted Code" },
	{ "S-1-5-13",				"Terminal Server Users" },
	{ "S-1-5-14",				"Remote Interactive Logon" },
	{ "S-1-5-15",				"This Organization" },
	{ "S-1-5-17",				"This Organization" },
	{ "S-1-5-18",				"Local System" },
	{ "S-1-5-19",				"NT Authority" },
	{ "S-1-5-20",				"NT Authority" },
	{ "S-1-5-32-544",			"Administrators" },
	{ "S-1-5-32-545",			"Users" },
	{ "S-1-5-32-546",			"Guests" },
	{ "S-1-5-32-547",			"Power Users" },
	{ "S-1-5-32-548",			"Account Operators" },
	{ "S-1-5-32-549",			"Server Operators" },
	{ "S-1-5-32-550",			"Print Operators" },
	{ "S-1-5-32-551",			"Backup Operators" },
	{ "S-1-5-32-552",			"Replicators" },
	{ "S-1-5-32-554",			"builtin\Pre-Windows 2000 Compatible Access" },
	{ "S-1-5-32-555",			"builtin\Remote Desktop Users" },
	{ "S-1-5-32-556",			"builtin\Network Configuration Operators" },
	{ "S-1-5-32-557",			"builtin\Incoming Forest Trust Builders" },
	{ "S-1-5-32-558",			"builtin\Performance Monitor Users" },
	{ "S-1-5-32-559",			"builtin\Performance Log Users" },
	{ "S-1-5-32-560",			"builtin\Windows Authorization Access Group" },
	{ "S-1-5-32-561",			"builtin\Terminal Server License Servers" },
	{ "S-1-5-32-562",			"builtin\Distributed COM Users" },
	{ "S-1-5-32-569",			"builtin\Cryptographic Operators" },
	{ "S-1-5-32-573",			"builtin\Event Log Readers" },
	{ "S-1-5-32-574",			"builtin\Certificate Service DCOM Access" },
	{ "S-1-5-32-575",			"builtin\RDS Remote Access Servers" },
	{ "S-1-5-32-576",			"builtin\RDS Endpoint Servers" },
	{ "S-1-5-32-577",			"builtin\RDS Management Servers" },
	{ "S-1-5-32-578",			"builtin\Hyper-V Administrators" },
	{ "S-1-5-32-579",			"builtin\Access Control Assistance Operators" },
	{ "S-1-5-32-580",			"builtin\Remote Management Users" },
	{ "S-1-5-64-10",			"NTLM Authentication" },
	{ "S-1-5-64-14",			"SChannel Authentication" },
	{ "S-1-5-64-21",			"Digest Authenitication" },
	{ "S-1-5-64-1000",			"Other Organization" },
	{ "S-1-5-80",				"NT Service" },
	{ "S-1-5-80-0",				"All Services" },
	{ "S-1-5-83-0",				"NT VIRTUAL MACHINE\Virtual Machines" },
	{ "S-1-6",					"Site Server Authority An identifier authority" },
	{ "S-1-7",					"Internet Site Authority An identifier authority" },
	{ "S-1-8",					"Exchange Authority An identifier authority" },
	{ "S-1-9",					"Resource Manager Authority An identifier" },
	{ "S-1-16-0",				"Untrusted Mandatory Level" },
	{ "S-1-16-4096",			"Low Mandatory Level" },
	{ "S-1-16-8192",			"Medium Mandatory Level" },
	{ "S-1-16-8448",			"Medium Plus Mandatory Level" },
	{ "S-1-16-12288",			"High Mandatory Level" },
	{ "S-1-16-16384",			"System Mandatory Level" },
	{ "S-1-16-20480",			"Protected Process Mandatory Level" },
	{ "S-1-16-28672",			"Secure Process Mandatory Level" },
};

static Num2Str GroupRids[] = {		/* Rid to Sid mapping:  "S-1-5-21<domain ID>-<rid>" */
	{ 498,			"Enterprise Read-only Domain Controllers" },
	{ 500,			"Administrator" },
	{ 501,			"Guest" },
	{ 502,			"KRBTGT" },
	{ 512,			"Domain Admins" },
	{ 513,			"Domain Users" },
	{ 514,			"Domain Guests" },
	{ 515,			"Domain Computers" },
	{ 516,			"Domain Controllers" },
	{ 517,			"Cert Publishers" },
	{ 518,			"Cert Publishers" },
	{ 519,			"Schema Admins" },
	{ 520,			"Group Policy Creator Owners" },
	{ 521,			"Read-only Domain Controllers" },
	{ 522,			"Cloneable Domain Controllers" },
	{ 526,			"Key Admins" },
	{ 527,			"Enterprise Key Admins" },
	{ 553,			"RAS and IAS Servers" },
	{ 571,			"Allowed RODC Password Replication Group" },
	{ 572,			"Denied RODC Password Replication Group" },
};

static Str2Str Oids[] = {
/* Microsoft Active Directory extensions */
	{ "1.2.840.113556.1.4.319",			"PAGED_RESULT_STRING" },
	{ "1.2.840.113556.1.4.521",			"SERVER_CROSSDOM_MOVE_TARGET" },
	{ "1.2.840.113556.1.4.841",			"SERVER_DIRSYNC" },
	{ "1.2.840.113556.1.4.1339",		"SERVER_DOMAIN_SCOPE" },
	{ "1.2.840.113556.1.4.529",			"SERVER_EXTENDED_DN" },
	{ "1.2.840.113556.1.4.970",			"SERVER_GET_STATS" },
	{ "1.2.840.113556.1.4.619",			"SERVER_LAZY_COMMIT" },
	{ "1.2.840.113556.1.4.1413",		"SERVER_PERMISSIVE_MODIFY" },
	{ "1.2.840.113556.1.4.528",			"SERVER_NOTIFICATION" },
	{ "1.2.840.113556.1.4.474",			"SERVER_RESP_SORT" },
	{ "1.2.840.113556.1.4.801",			"SERVER_SD_FLAGS" },
	{ "1.2.840.113556.1.4.1340",		"SERVER_SEARCH_OPTIONS" },
	{ "1.2.840.113556.1.4.473",			"SERVER_SORT" },
	{ "1.2.840.113556.1.4.417",			"SERVER_SHOW_DELETED" },
	{ "1.2.840.113556.1.4.805",			"SERVER_TREE_DELETE" },
	{ "1.2.840.113556.1.4.1338",		"SERVER_VERIFY_NAME" },
	{ "2.16.840.1.113730.3.4.9",		"CONTROL_VLVREQUEST" },
	{ "2.16.840.1.113730.3.4.10",		"CONTROL_VLVRESPONSE" },
	{ "1.2.840.113556.1.4.1504",		"SERVER_ASQ" },
	{ "1.2.840.113556.1.4.1852",		"SERVER_QUOTA_CONTROL" },
	{ "1.2.840.113556.1.4.802",			"SERVER_RANGE_OPTION" },
	{ "1.2.840.113556.1.4.1907",		"SERVER_SHUTDOWN_NOTIFY" },
	{ "1.2.840.113556.1.4.1974",		"SERVER_FORCE_UPDATE" },
	{ "1.2.840.113556.1.4.1948",		"SERVER_RANGE_RETRIEVAL_NOERR" },
	{ "1.2.840.113556.1.4.1341",		"SERVER_RODC_DCPROMO" },
	{ "1.2.840.113556.1.4.2026",		"SERVER_DN_INPUT" },
	{ "1.2.840.113556.1.4.2065",		"SERVER_SHOW_DEACTIVATED_LINK" },
	{ "1.2.840.113556.1.4.2064",		"SERVER_SHOW_RECYCLED" },
	{ "1.2.840.113556.1.4.2066",		"SERVER_POLICY_HINTS_DEPRECATED" },
	{ "1.2.840.113556.1.4.2090",		"SERVER_DIRSYNC_EX" },
	{ "1.2.840.113556.1.4.2205",		"SERVER_UPDATE_STATS" },
	{ "1.2.840.113556.1.4.2204",		"SERVER_TREE_DELETE_EX" },
	{ "1.2.840.113556.1.4.2206",		"SERVER_SEARCH_HINTS" },
	{ "1.2.840.113556.1.4.2211",		"SERVER_EXPECTED_ENTRY_COUNT" },
	{ "1.2.840.113556.1.4.2239",		"SERVER_POLICY_HINTS" },
	{ "1.2.840.113556.1.4.2255",		"SERVER_SET_OWNER" },
	{ "1.2.840.113556.1.4.2256",		"SERVER_BYPASS_QUOTA" },
	{ "1.2.840.113556.1.4.2309",		"SERVER_LINK_TTL" },
	{ "1.2.840.113556.1.4.2330",		"SERVER_SET_CORRELATION_ID" },
	{ "1.2.840.113556.1.4.2354",		"SERVER_THREAD_TRACE_OVERRIDE" },

/* Novell extensions */
	{ "1.3.6.1.4.1.1466.20037",				"Start TLS" },

	{ "2.16.840.1.113719.1.27.100.1",		"ndsToLdapResponse" },
	{ "2.16.840.1.113719.1.27.100.2",		"ndsToLdapRequest" },
	{ "2.16.840.1.113719.1.27.100.3",		"createNamingContextRequest" },
	{ "2.16.840.1.113719.1.27.100.4",		"createNamingContextResponse" },
	{ "2.16.840.1.113719.1.27.100.5",		"mergeNamingContextRequest" },
	{ "2.16.840.1.113719.1.27.100.6",		"mergeNamingContextResponse" },
	{ "2.16.840.1.113719.1.27.100.7",		"addReplicaRequest" },
	{ "2.16.840.1.113719.1.27.100.8",		"addReplicaResponse" },
	{ "2.16.840.1.113719.1.27.100.9",		"refreshLDAPServerRequest" },
	{ "2.16.840.1.113719.1.27.100.10",		"refreshLDAPServerResponse" },
	{ "2.16.840.1.113719.1.27.100.11",		"removeReplicaRequest" },
	{ "2.16.840.1.113719.1.27.100.12",		"removeReplicaResponse" },
	{ "2.16.840.1.113719.1.27.100.13",		"namingContextEntryCountRequest" },
	{ "2.16.840.1.113719.1.27.100.14",		"namingContextEntryCountResponse" },
	{ "2.16.840.1.113719.1.27.100.15",		"changeReplicaTypeRequest" },
	{ "2.16.840.1.113719.1.27.100.16",		"changeReplicaTypeResponse" },
	{ "2.16.840.1.113719.1.27.100.17",		"getReplicaInfoRequest" },
	{ "2.16.840.1.113719.1.27.100.18",		"getReplicaInfoResponse" },
	{ "2.16.840.1.113719.1.27.100.19",		"listReplicaRequest" },
	{ "2.16.840.1.113719.1.27.100.20",		"listReplicaResponse" },
	{ "2.16.840.1.113719.1.27.100.21",		"receiveAllUpdatesRequest" },
	{ "2.16.840.1.113719.1.27.100.22",		"receiveAllUpdatesResponse" },
	{ "2.16.840.1.113719.1.27.100.23",		"sendAllUpdatesRequest" },
	{ "2.16.840.1.113719.1.27.100.24",		"sendAllUpdatesResponse" },
	{ "2.16.840.1.113719.1.27.100.25",		"requestNamingContextSyncRequest" },
	{ "2.16.840.1.113719.1.27.100.26",		"requestNamingContextSyncResponse" },
	{ "2.16.840.1.113719.1.27.100.27",		"requestSchemaSyncRequest" },
	{ "2.16.840.1.113719.1.27.100.28",		"requestSchemaSyncResponse" },
	{ "2.16.840.1.113719.1.27.100.29",		"abortNamingContextOperationRequest" },
	{ "2.16.840.1.113719.1.27.100.30",		"abortNamingContextOperationResponse" },
	{ "2.16.840.1.113719.1.27.100.31",		"getContextIdentityNameRequest" },
	{ "2.16.840.1.113719.1.27.100.32",		"getContextIdentityNameResponse" },
	{ "2.16.840.1.113719.1.27.100.33",		"getEffectivePrivilegesRequest" },
	{ "2.16.840.1.113719.1.27.100.34",		"getEffectivePrivilegesResponse" },
	{ "2.16.840.1.113719.1.27.100.35",		"SetReplicationFilterRequest" },
	{ "2.16.840.1.113719.1.27.100.37",		"getReplicationFilterRequest" },
	{ "2.16.840.1.113719.1.27.100.39",		"createOrphanPartitionrequest" },
	{ "2.16.840.1.113719.1.27.100.41",		"removeOrphanPartitionRequest" },
	{ "2.16.840.1.113719.1.27.100.43",		"triggerBKLinkerRequest" },
	{ "2.16.840.1.113719.1.27.100.45",		"triggerDRLProcessRequest" },
	{ "2.16.840.1.113719.1.27.100.47",		"triggerJanitorRequest" },
	{ "2.16.840.1.113719.1.27.100.49",		"triggerLimberRequest" },
	{ "2.16.840.1.113719.1.27.100.51",		"triggerSkulkerRequest" },
	{ "2.16.840.1.113719.1.27.100.53",		"triggerSchemaSyncRequest" },
	{ "2.16.840.1.113719.1.27.100.55",		"triggerPartitionPurgeRequest" },
	{ "2.16.840.1.113719.1.27.100.79",		"EventMonitorRequest" },
	{ "2.16.840.1.113719.1.27.100.84",		"filteredEventMonitorRequest" },

	{ "2.16.840.1.113719.1.27.103.1",		"createGroupingRequest" },
	{ "2.16.840.1.113719.1.27.103.2",		"endGroupingRequest" },

	{ "2.16.840.1.113719.1.148.100.1",		"Put Login Configuration" },
	{ "2.16.840.1.113719.1.148.100.3",		"Get Login Configuration" },
	{ "2.16.840.1.113719.1.148.100.5",		"Delete Login Configuration" },
	{ "2.16.840.1.113719.1.148.100.7",		"Put Login Secret" },
	{ "2.16.840.1.113719.1.148.100.9",		"Delete Login Secret" },
	{ "2.16.840.1.113719.1.148.100.11",		"Set Universal Password" },
	{ "2.16.840.1.113719.1.148.100.13",		"Get Universal Password" },
	{ "2.16.840.1.113719.1.148.100.15",		"Delete Universal Password" },
	{ "2.16.840.1.113719.1.148.100.17",		"Check password against password policy" },

	{ "2.16.840.1.113719.1.39.42.100.19",	"Get password policy information" },
	{ "2.16.840.1.113719.1.39.42.100.21",	"Change Universal Password" },
	{ "2.16.840.1.113719.1.39.42.100.23",	"Graded Authentication management" },
	{ "2.16.840.1.113719.1.39.42.100.25",	"NMAS management (new with NMAS 3.1.0)" },

	{ "2.16.840.1.113719.1.148.100.1",		"SSLDAP_GET_SERVICE_INFO_REQUEST" },
	{ "2.16.840.1.113719.1.148.100.2",		"SSLDAP_GET_SERVICE_INFO_REPLY" },
	{ "2.16.840.1.113719.1.148.100.3",		"SSLDAP_READ_SECRET_REQUEST" },
	{ "2.16.840.1.113719.1.148.100.4",		"SSLDAP_READ_SECRET_REPLY" },
	{ "2.16.840.1.113719.1.148.100.5",		"SSLDAP_WRITE_SECRET_REQUEST" },
	{ "2.16.840.1.113719.1.148.100.6",		"SSLDAP_WRITE_SECRET_REPLY" },
	{ "2.16.840.1.113719.1.148.100.7",		"SSLDAP_ADD_SECRET_ID_REQUEST" },
	{ "2.16.840.1.113719.1.148.100.8",		"SSLDAP_ADD_SECRET_ID_REPLY" },
	{ "2.16.840.1.113719.1.148.100.9",		"SSLDAP_REMOVE_SECRET_REQUEST" },
	{ "2.16.840.1.113719.1.148.100.10",		"SSLDAP_REMOVE_SECRET_REPLY" },
	{ "2.16.840.1.113719.1.148.100.11",		"SSLDAP_REMOVE_SECRET_STORE_REQUEST" },
	{ "2.16.840.1.113719.1.148.100.12",		"SSLDAP_REMOVE_SECRET_STORE_REPLY" },
	{ "2.16.840.1.113719.1.148.100.13",		"SSLDAP_ENUMERATE_SECRET_IDS_REQUEST" },
	{ "2.16.840.1.113719.1.148.100.14",		"SSLDAP_ENUMERATE_SECRET_IDS_REPLY" },
	{ "2.16.840.1.113719.1.148.100.15",		"SSLDAP_UNLOCK_SECRETS_REQUEST" },
	{ "2.16.840.1.113719.1.148.100.16",		"SSLDAP_UNLOCK_SECRETS_REPLY" },
	{ "2.16.840.1.113719.1.148.100.17",		"SSLDAP_SET_EP_MASTER_PASSWORD_REQUEST" },
	{ "2.16.840.1.113719.1.148.100.18",		"SSLDAP_SET_EP_MASTER_PASSWORD_REPLY" },

	{ "2.16.840.1.113719.1.27.100.103",		"Get Privileges List Request" },
	{ "2.16.840.1.113719.1.27.100.104",		"Get Privileges List Response" },

};

static Num2Str Errors[] = {
	{ 0,		"Ok" },
	{ 1,		"request out of order" },
	{ 2,		"bad request" },
	{ 3,		"Timeout servicing request" },
	{ 4,		"Result too big" },
	{ 5,		"compare false" },
	{ 6,		"compare true" },
	{ 7,		"auth method not supported" },
	{ 8,		"strong auth required" },
	{ 9,		"Reserved." },
	{ 10,		"referal" },
	{ 11,		"admin limit exceeded" },
	{ 12,		"required server extension not available" },
	{ 13,		"confidentiality required" },
	{ 14,		"SASL bind in progress" },
	{ 15,		"not used." },
	{ 16,		"no such attribute" },
	{ 17,		"undefined type" },
	{ 18,		"inappropriate matching" },
	{ 19,		"constraint violation" },
	{ 20,		"type or attribute exists" },
	{ 21,		"invalid syntax" },
	{ 32,		"no such object" },
	{ 33,		"aliais problem" },
	{ 34,		"invalid DN= syntax" },
	{ 35,		"is a leaf node" },
	{ 36,		"alisa deref problem" },
	{ 48,		"inappropriate auth" },
	{ 49,		"invalid credentials" },
	{ 50,		"insufficent access rights" },
	{ 51,		"server busy" },
	{ 52,		"unavailable" },
	{ 53,		"unwilling to peform" },
	{ 54,		"loop detected" },
	{ 64,		"naming violation" },
	{ 65,		"object class violation" },
	{ 66,		"not allowed on a leaf node" },
	{ 67,		"not allowed on RDN=" },
	{ 68,		"already exists" },
	{ 69,		"cannot modify object class" },
	{ 70,		"result too large" },
	{ 71,		"cross server move" },
	{ 80,		"unknown error" },
};

static Num2Str GroupTypes[] = {
	{ 0x00000001,		"created by the system"},
	{ 0x00000002,		"global scope"},
	{ 0x00000004,		"domain scope"},
	{ 0x00000008,		"universal scope"},
	{ 0x00000010,		"Basic group for SAM"},
	{ 0x00000020,		"Query group for SAM"},
	{ 0x80000000,		"security group"},
};

static Num2Str SystemFlags[] = {
	{ 0x00000001,		"not replicated / context in NTDS" },
	{ 0x00000002,		"in global catalog / context is a domain" },
	{ 0x00000004,		"is constructed" },
	{ 0x00000010,		"category 1 object" },
	{ 0x02000000,		"do not detete to the Deleted Objects" },
	{ 0x04000000,		"cannot be moved" },
	{ 0x08000000,		"cannot be renamed" },
	{ 0x10000000,		"move limited" },
	{ 0x20000000,		"limited move" },
	{ 0x40000000,		"rename limited" },
	{ 0x80000000,		"readonly" },
};

static Num2Str InstanceType[] = {
	{ 0x01,	"head of naming context" },
	{ 0x02,	"replica not instantiated" },
	{ 0x04,	"readonly" },
	{ 0x08,	"have parent" },
	{ 0x10,	"replication in progress" },
	{ 0x20,	"delete in progress" },
};

static Num2Str Boolean[] = {
	{ 0,	"false" },
	{ 1,	"true" },
};

/**********************************/

/* get a large integer */
static vlong
by2ll(Bytes *b)
{
	int n;
	char str[64];

	n = sizeof(str);
	if(n > b->len)
		n = b->len;
	memcpy(str, b->data, n);
	str[n] = 0;
	return atoll(str);
}

/* return true if ths param's name matches (case independent) */
static int
bycmp(Bytes *b, char *s)
{
	return cistrncmp(s, (char *)(b->data), b->len);
}

/**********************************/

static int
_numtab(Fmt *f, Bytes *b, int num, void *aux)
{
	int i;
	uvlong n;
	Num2Str *ns;

	ns = aux;
	n = by2ll(b);
	for(i = 0; i < num; i++)
		if(n == ns[i].n)
			return fmtprint(f, "%s", ns[i].s);
	return fmtprint(f, "<%B>", b);
}

static int
_strtab(Fmt *f, Bytes *b, int num, void *aux)
{
	int i, rc;
	Str2Str *ss;

	ss = aux;
	for(i = 0; i < num; i++)
		if(bycmp(b, ss[i].x) == 0)
			return fmtprint(f, "%s", ss[i].y);
	rc = fmtprint(f, "<%B>", b);
	return rc;
}

static int
_bitmask(Fmt *f, Bytes *b, int num, void *aux)
{
	uvlong n;
	Num2Str *ns;
	int i, rc;
	char *prefix;

	rc = 0;
	ns = aux;
	prefix = "";
	n = by2ll(b);
	for(i = 0; i < num; i++){
		if(ns[i].n & n){
			rc += fmtprint(f, "%s%s", prefix, ns[i].s);
			prefix = ", ";
		}
	}
	return rc;
}

static int
_security_id(Fmt *f, Bytes *b, int, void *)
{
	ulong n;
	uchar *p;
	char buf[1024];
	int i, j, num, rev, len;

	if(b->len < 8){
		werrstr("sid too short (%d < 8)", b->len);
		return -1;
	}

	if(b->len < 8 || (b->len % 4) != 0){
		werrstr("bad sid length");
		return -1;
	}

	p = b->data;

	rev = *p++;
	if(rev != 1){
		werrstr("SID revision not supported (%d != 1)", rev);
		return -1;
	}

	num = *p++;
	if(num > 15){
		werrstr("too many sub authorities in SID (%d > 15)", num);
		return -1;
	}

	snprint(buf, sizeof(buf), "S-%d", rev);
	len = strlen(buf);

	n = 0;
	for(j = 0; j < 6; j++)			/* identifier authority is 48 bits is big endian */
		n = (n<<8) | *p++;
	snprint(buf+len, sizeof(buf)-len, "-%lud", n);
	len = strlen(buf);

	for(i = 0; i < num; i++){		/* sub authorities are 32 bits little endian */
		n = 0;
		for(j = 0; j < 4; j++)
			n = n | (*p++<<(8*j));
		snprint(buf+len, sizeof(buf)-len, "-%lud", n);
		len = strlen(buf);
	}

	for(i = 0; i < nelem(WellKnownSids); i++)
		if(strcmp(WellKnownSids[i].x, buf) == 0)
			return fmtprint(f, "%s", WellKnownSids[i].y);

	return fmtprint(f, "%s", buf);
}

static int
_globally_unique_id(Fmt *f, Bytes *b, int, void *)
{
	uchar *p;
	int rc, i;

	if(b->len != 16){
		werrstr("bad guid length");
		return -1;
	}

	rc = 0;
	p = b->data;
	rc += fmtprint(f, "{");
	for(i = 0; i < 4; i++)
		rc += fmtprint(f, "%02x", *p++);
	rc += fmtprint(f, "-");
	for(i = 0; i < 2; i++)
		rc += fmtprint(f, "%02x", *p++);
	rc += fmtprint(f, "-");
	for(i = 0; i < 2; i++)
		rc += fmtprint(f, "%02x", *p++);
	rc += fmtprint(f, "-");
	for(i = 0; i < 2; i++)
		rc += fmtprint(f, "%02x", *p++);
	rc += fmtprint(f, "-");
	for(i = 0; i < 6; i++)
		rc += fmtprint(f, "%02x", *p++);
	rc += fmtprint(f, "}");
	return rc;
}


static int
_abs_time(Fmt *f, Bytes *b, int, void *)
{
	uvlong t;
	char *s, *p;

	t = mswin2epoch(b);
	if(t <= 0)
		return fmtprint(f, "never");

	s = ctime(t);
	if((p = strchr(s, '\n')) != nil)
		*p = 0;
	return fmtprint(f, "%s", s);
}

static int
_elapsed_time(Fmt *f, Bytes *b, int, void *)
{
	long d;
	long t;
	char *p, *s;

	t = mswin2epoch(b);
	if(t <= 0)
		return fmtprint(f, "never");

	d = (t - time(nil)) / (60 * 60 * 24);

	if(d == 0)
		return fmtprint(f, "now");

	if(d > 0 && d < 300)
		return fmtprint(f, "in %ld days", d);

	if(d < 0 && d > -300)
		return fmtprint(f, "%ld days ago", -d);

	s = ctime(t);
	if((p = strchr(s, '\n')) != nil)
		*p = 0;
	return fmtprint(f, "%s", s);
}

/*
 *	"createTimestamp"
 *	"modifyTimestamp"
 */
static int
_iso_time(Fmt *f, Bytes *b, int, void *)
{
	long t;
	char *s, *p;

	if((t = iso2epoch(b)) <= 0)
		return fmtprint(f, "never");

	s = ctime(t);
	if((p = strchr(s, '\n')) != nil)
		*p = 0;
	return fmtprint(f, "%s", s);
}


/*
 *	protocolSettings
 */
static int
_proto_set(Fmt *f, Bytes *b, int, void *)
{
	Rune r;
	char *s, *p;
	int ch, field;

	ch = 0;
	field = 1;
	p = (char *)b->data;
	do{
		p += chartorune(&r, p);
		switch(r){
		case L'0':
			s = (field == 2)?  "disabled": "0";
			ch += fmtprint(f, "%s", s);
			break;
		case L'1':
			s = (field == 2)?  "enabled": "0";
			ch += fmtprint(f, "%s", s);
			break;
		case 0:
			break;
		case L'§':
			s = (field == 1)?  "=": ", ";
			ch += fmtprint(f, "%s", s);
			field++;
			break;
		default:
			ch += fmtprint(f, "%C", r);
			break;
		}
	}while(r != 0);
	return ch;
}


static int
_hexdump(Fmt *f, Bytes *b, int, void *)
{
	int i, first, rc;

	rc = 0;
	first = 1;
	for(i = 0; i < b->len; i++){
		if(! first)
			fmtprint(f, " ");
		rc += fmtprint(f, "%02x", b->data[i]);
		first = 0;
	}
	return rc;
}

/**********************************************/

/*
 *  user access control,
 * see http://www.selfadsi.org/ads-attributes/user-userAccountControl.htm
 */
int
fmt_namevalue(Fmt *f)
{
	int i;
	Namval *nv;

	static struct {
		char *name;
		int (*func)(Fmt *, Bytes *, int, void *);
		int num;
		void *aux;
	} Formats[] = {
		{ "",									_hexdump },
		{ "protocolSettings",					_proto_set },
		{ "msExchSafeSendersHash",				_hexdump },
		{ "msExchMailboxSecurityDescriptor",	_hexdump },
		{ "lockoutDuration",					_elapsed_time },
		{ "lockOutObservationWindow",			_elapsed_time },
		{ "maxPwdAge",							_elapsed_time },
		{ "netbootGUID",						_globally_unique_id },
		{ "objectGUID",							_globally_unique_id },
		{ "msExchMailboxGuid",					_globally_unique_id },
		{ "msExchMasterAccountSid",				_security_id },
		{ "objectSid",							_security_id },
		{ "mS-DS-CreatorSID",					_security_id },
		{ "badPasswordTime",					_abs_time },
		{ "lastLogon",							_abs_time },
		{ "pwdLastSet",							_abs_time },
		{ "creationTime",						_abs_time },
		{ "accountExpires",						_abs_time },
		{ "msExchWhenMailboxCreated",			_abs_time },
		{ "lastLogonTimestamp",					_abs_time },
		{ "whenCreated",						_iso_time },
		{ "dSCorePropagationData",				_iso_time },
		{ "whenCreated",						_iso_time },
		{ "modifyTimeStamp",					_iso_time },
		{ "whenChanged",						_iso_time },
		{ "instanceType",						_bitmask,		nelem(InstanceType), InstanceType },
		{ "primaryGroupID",						_numtab,		nelem(GroupRids), GroupRids },
		{ "msExchVersion",						_numtab,		nelem(ExchangeVersion), ExchangeVersion },
		{ "userAccountControl",					_bitmask,		nelem(AccountControl), AccountControl },
		{ "sAMAccountType",						_numtab,		nelem(AccountType), AccountType },
		{ "supportedControl",					_strtab,		nelem(Oids), Oids },
		{ "supportedCapabilities",				_strtab,		nelem(Oids), Oids },
		{ "uASCompat",							_numtab,		nelem(Boolean), Boolean  },
		{ "groupType",							_bitmask,		nelem(GroupTypes), GroupTypes  },
		{ "systemFlags",						_bitmask,		nelem(SystemFlags), SystemFlags  },
		{ "DNSRecord",							dnsRecord },


	/* format not known, these are only guesses */
		{ "msExchRecipientDisplayType",			_numtab, nelem(RecipentType),  RecipentType },
		{ "msExchBlockedSendersHash",			_hexdump },
	};

	nv = va_arg(f->args, Namval *);
	for(i = 0; i < nelem(Formats); i++)
		if(cistrncmp(Formats[i].name, (char *)nv->name->data, nv->name->len) == 0)
			return Formats[i].func(f, nv->value, Formats[i].num, Formats[i].aux);

	for(i = 0; i < nv->value->len; i++)
		if(nv->value->data[i] > 0x7f){
			fprint(2, "%s: %B non-ascii value\n", argv0, nv->name);
			return _hexdump(f, nv->value, 0, nil);
		}
	return fmtprint(f, "%B", nv->value);
}


int
fmt_text(Fmt *f)
{
	Bytes *b;

	b = va_arg(f->args, Bytes *);
	return fmtprint(f, "%.*s", b->len, (char *)b->data);
}

int
fmt_errstr(Fmt *f)
{
	int i, n;

	n = va_arg(f->args, int);
	for(i = 0; i < nelem(Errors); i++)
		if(Errors[i].n == n && Errors[i].s != nil)
			return fmtprint(f, "%s", Errors[i].s);

	return fmtprint(f, "%d - unknown LDAP error", n);
}

