#include <u.h>
#include <libc.h>
#include <auth.h>
#include <mp.h>
#include <libsec.h>
#include <der.h>
#include <fcall.h>
#include <thread.h>
#include <9p.h>
#include "ldap.h"

struct {
	int len;
	int data[7];
} PagedResults = {
	7, { 1, 2, 840, 113556, 1, 4, 319 },
};

/* send an ldap request */
static int
request(Ldap *l, int api, int num, Bytes *resume, Elist *lst)
{
	Elem e;
	Bytes *buf;
	int err, len, sent;

resume = nil;		/* FIXME: the resume encoding is broken at present */

	if(resume){
		e = mkseq(
			mkel(mkint(num),
			mkel(mkappl(api, lst),
			mkel(
				mkcontseq(0,
					mkel(
						mkseq(
							mkel(mkoid((Ints *)&PagedResults),
							mkel(mkbool(0),
							mkel(mkoctet(resume->data, resume->len),
							nil
						)))), nil)), nil))));
	}
	else{
		e = mkseq(
			mkel(mkint(num),
			mkel(mkappl(api, lst),
			nil
			)));
	}


	if((err = der_encode(e, &buf)) != ASN_OK){
		werrstr("der_encode: %s", Asn1Errstr[err]);
		return -1;
	}

	freevalfields(&e.val);

	len = buf->len;
	sent = write(l->fd, buf->data, buf->len);

	if(Debug && strstr(Debug, "der")){
		fprint(2, "tx sent=%d\n", sent);
		if(sent > 0)
			der_dump(buf->data, sent);
	}
	freebytes(buf);

	if(sent < len)
		return -1;
	return 0;
}

/* calculate how much more data will arrive given the header */
static int
how_much_more(uchar *buf, int len)
{
	int i, n, size;

	if((buf[0] & CONSTR_MASK) != CONSTR_MASK){
		werrstr("reply does not begin with a CONSTRUCTED object");
		return -1;
	}

	if((buf[1] & 0x80) == 0)				/* size is less than 128 */
		return (buf[1] +2) -len;

	n = buf[1] & 0x7f;
	if(n == 0){								/* terminator as length not used in DER */
		werrstr("reply SEQUENCE length is inferred (not supported)");
		der_dump(buf, len);
		return -1;
	}

	size = 0; 
	for(i = 0; i < n && i < len; i++)		/* size is greater than 127 */
		size = (size << 8) | buf[i+2];
	return (size + 2 + n) - len;
}

/* await and process a reply from the server */
static int
answer(Ldap *l, int maxmsg, Elem *ep)
{
	Bytes *buf;
	uchar hdr[8];
	int got, need, n, err;

	alarm(10000);
	got = readn(l->fd, hdr, sizeof(hdr));
	alarm(0);
	if(got < 0){
		werrstr("read header failed, %r");
		return -1;
	}

	if((need = how_much_more(hdr, got)) < 0)
		return -1;

	buf = newbytes(maxmsg);
	memcpy(buf->data, hdr, got);

	if(need != 0){
		alarm(1000);
		n = readn(l->fd, buf->data+got, need);
		alarm(0);
		if(n < 0){
			werrstr("read body failed, %r");
			return -1;
		}

		if(n != need){
			werrstr("bad message length, (%d != %d), %r", n, need);
			freebytes(buf);
			return -1;
		}
		got += n;
	}

	if(Debug && strstr(Debug, "der")){
		fprint(2, "rx got=%d\n", got);
		der_dump(buf->data, got);
	}

	if((err = der_decode(buf->data, got, ep)) != ASN_OK){
		werrstr("der_decode: %s", Asn1Errstr[err]);
		freebytes(buf);
		return -1;
	}

	freebytes(buf);
	return 0;
}

/* bump and wrap sequence number */
static int
seqno(Ldap *l)
{
	int n;

	n = l->seqno++;
	if(n <= 0)			/* NB: seq of zero is reserved */
		n = 1;
	return n;
}

Elem
mkoctstr(char *str)
{
	return mkoctet((uchar *)str, strlen(str));
}

/* connect (dial and start TLS) to the server */
Ldap *
ldap_connect(char *server)
{
	int fd;
	Ldap *l;
	Thumbprint *thumb;
	uchar digest[SHA1dlen];
	char *addr, *host, *port;

	fmtinstall('H', encodefmt);

	if((addr = netmkaddr(server, "tcp", "ldap")) == nil)
		sysfatal("%s bad address, %r", server);
	if((fd = dial(addr, "ldap", nil, nil)) == -1)
		sysfatal("%s dial failed, %r", addr);

	if((l = emalloc9p(sizeof(Ldap))) == nil)
		return nil;

	l->fd = fd;
	l->seqno = 1;							/* sequence number zero is reserved */
	l->host = server;

	if((host = strchr(addr, '!')) == nil)	/* no netdir */
		return l;
	*host++ = 0;
	l->host = host;

	if((port = strrchr(host, '!')) == nil)	/* no port */
		return l;
	*port++ = 0;
	l->port = port;
	
	if(strcmp(port, "ldaps") != 0)			/* not TLS port */
		return l;

	l->sec = 1;
	if((thumb = initThumbprints("/sys/lib/tls/ldap", "/sys/lib/tls/ldap.exclude", "x509")) == nil){
		free(l);
		close(fd);
		return nil;
	}

	if((l->fd = tlsClient(fd, &l->conn)) < 0){
		free(l);
		close(fd);
		return nil;
	}
	close(fd);

	if(thumb){
		if(l->conn.cert==nil || l->conn.certlen<=0)
			sysfatal("server did not provide TLS certificate");
		sha1(l->conn.cert, l->conn.certlen, digest, nil);
		if(! okThumbprint(digest, sizeof(digest), thumb))
			sysfatal("%s presented certificate '%.*H', not known or excluded", l->host, SHA1dlen, digest);
	}
	free(l->conn.cert);
	return l;
}

/* bind (authenticate) to the server */
int
ldap_bind(Ldap *l, char *user, char *passwd)
{
	Elem rx;
	Bytes *errmsg;
	Elist *tx, *hdr, *payload;
	int err, len, rxapi, txseq, rxseq;

	tx = mkel(mkint(ProtoVers),
			mkel(mkoctstr(user),
			mkel(mkcontstr(0, passwd),
			nil)));

	txseq = seqno(l);
	if((err = request(l, Tbind, txseq, nil, tx)) < 0)
		return err;
	if((err = answer(l, 1024, &rx)) < 0){
		return err;
	}
	if(!is_seq(&rx, &hdr)){								/* starts with a SEQUENCE? */
		werrstr("ldap reply - not a SEQUENCE\n");
		freevalfields(&rx.val);
		return -1;
	}

	if((len = elistlen(hdr)) != 2){						/* SEQUENCE length is correct (2)? */
		werrstr("ldap reply - bad header length (%d!=2)\n", len);
		freevalfields(&rx.val);
		return -1;
	}

	if(!is_int(&hdr->hd, &rxseq)){						/* is a sequence number? */
		werrstr("ldap reply - missing seqno\n");
		freevalfields(&rx.val);
		return -1;
	}

	if(! is_appl(&hdr->tl->hd, &rxapi, &payload)){		/* is an Application block? */
		werrstr("ldap reply - not an appl block\n");
		freevalfields(&rx.val);
		return -1;
	}

	if(rxapi != Rbind){									/* is a bind reply? */
		werrstr("incorrect reply (expeteced Rbind)");
		freevalfields(&rx.val);
		return -1;
	}

	if(rxseq != txseq){									/* message sequence number is correct? */
		werrstr("mismatched sequence numbers (%d != %d)", rxseq, txseq);
		freevalfields(&rx.val);
		return -1;
	}

	if(!is_int(&payload->hd, &err)){					/* have a result code? */
		werrstr("ldap reply - missing result code\n");
		freevalfields(&rx.val);
		return -1;
	}

	if(err > 0){										/* result is an error? */
		/*
		 * this is not very useful
		 * is_octetstring(&payload)->hd, &rxdn);
		 */

		if(!is_octetstring(&payload->tl->hd, &errmsg))
			sysfatal("ldap_bind bad reply format, no error text\n");
		werrstr("%R %B", err, errmsg);
		freevalfields(&rx.val);
		return -1;
	}

	freevalfields(&rx.val);
	return 0;
}


/* search results */
int
search_results(Elist *payload, Srch *srch)
{
	Namval nv;
	Elem reply;
	Bytes *dn;
	Elist *attrs, *a, *v, *pair, *vals;
	
	if(! is_octetstring(&payload->hd, &dn)){
		werrstr("missing dn\n");
		return -1;
	}

	if(Verbose)
		print(" object: dn=%B\n", dn);
	if(srch->object)
		srch->object(srch->aux, dn);

	reply = payload->tl->hd;
	if(!is_seq(&reply, &attrs)){
		werrstr("not a SEQUENCE\n");
		return -1;
	}
	if(attrs == nil){
		werrstr("empty SEQUENCE\n");
		return 0;
	}
		
	for(a = attrs; a; a = a->tl){				
		if(!is_seq(&a->hd, &pair)){
			fprint(2, "not a sub sequence\n");
			continue;
		}
		if(!is_octetstring(&pair->hd, &nv.name)){
			fprint(2, "field name not an octet string\n");
			continue;
		}
		if(!is_set(&pair->tl->hd, &vals)){
			fprint(2, "not a set\n");
			continue;
		}
		for(v = vals; v; v = v->tl){
			if(!is_octetstring(&v->hd, &nv.value)){
				fprint(2, "value not a octet string\n");
				continue;
			}
			if(Verbose)
				print(" attrib: '%B=%L'\n", nv.name, &nv);
			if(srch->attrib)
				srch->attrib(srch->aux, &nv);
		}
	}
	return 0;
}

/* search results - reference (url) to an alternative ldap server/dn */
static int
search_redirect(Elist *payload, Srch *srch)
{
	Bytes *url;

	if(!is_octetstring(&payload->hd, &url)){
		fprint(2, "ldap search: no url\n");
		return -1;
	}
	if(Verbose)
		print(" redirect: '%B'\n", url);
	if(srch->refer)
		srch->refer(srch->aux, url);
	return 0;
}

/* search finished */
static int
search_done(Elist *payload)
{
	int err;
	Bytes *match, *diag;

	diag = nil;

	if(!is_int(&payload->hd, &err)){
		werrstr("ldap search: missing error code\n");
		return -1;
	}
	
	if(err){
		if(!is_octetstring(&payload->tl->hd, &match)){
			fprint(2, "ldap search: no match dn\n");
			return -1;
		}

		if(!is_octetstring(&payload->tl->tl->hd, &diag)){
			fprint(2, "ldap search: no match dn\n");
			return -1;
		}
		werrstr("%R, %B", err, diag);
		if(Verbose)
			fprint(2, " error: '%R' '%B'\n", err, diag);
		return -1;
	}
	if(Verbose)
		print(" done\n");
	return 0;
}

/* search request to server */
int
ldap_search(Ldap *l, char *dn, Elem attrs, int scope, Elem filter, int results, Srch *srch)
{
	Elem rx;
	Bytes empty;
	Elist *tx, *hdr, *payload;
	int err, rxapi, txseq, rxseq;

	if(Verbose)
		print(" search dn=%s results=%d\n", dn, results);

	if(Debug && strstr(Debug, "all"))			/* get all attributes */
		attrs = mkseq(mkel(mkoctstr("*"), nil));

	tx = mkel(mkoctstr(dn),						/* base DN= */
			mkel(mkenum(scope),					/* search depth */
			mkel(mkenum(DerefAlways),			/* when to de-reference results */
			mkel(mkint(results),				/* max number of entries to return (0=unlimited) */
			mkel(mkint(0),						/* max number of seconds to search for (0=unlimited) */
			mkel(mkbool(TypesAndData),			/* return types only or types and data */
			mkel(filter,
			mkel(attrs, 
			nil))))))));

	empty.len = 0;
	txseq = seqno(l);

	if((err = request(l, Tsearch, txseq, &empty, tx)) < 0)
		return err;

	err = 0;
	USED(err);
	do{
		if((err = answer(l, 0xffff, &rx)) < 0)
			return err;

		if(!is_seq(&rx, &hdr)){								/* starts with a SEQUENCE? */
			werrstr("ldap reply - not a SEQUENCE\n");
			freevalfields(&rx.val);
			return -1;
		}
	
		if(!is_int(&hdr->hd, &rxseq)){						/* is a sequence number? */
			werrstr("ldap search: missing seqno\n");
			freevalfields(&rx.val);
			return -1;
		}
	
		if(! is_appl(&hdr->tl->hd, &rxapi, &payload)){		/* is an Application block? */
			werrstr("ldap search: not an appl block\n");
			freevalfields(&rx.val);
			return -1;
		}
	
		if(rxseq != txseq){									/* message sequence number is correct? */
			werrstr("mismatched sequence numbers (%d != %d)", rxseq, txseq);
			freevalfields(&rx.val);
			return -1;
		}

		switch(rxapi){
		case RsearchEntry:				/* search reply */
			search_results(payload, srch);
			break;
		case Rreference:				/* redirect to URL */
			search_redirect(payload, srch);
			break;
		case Rsearch:					/* search completed */
			err |= search_done(payload);
			break;
		}

		freevalfields(&rx.val);
	}while(rxapi != Rsearch);

	return err;
}

/* cancel an ongoing request */
int
ldap_cancel(Ldap *l, int id)
{
	return request(l, Tabandon, id, nil, nil);	
}

/* unbind from server (may re-bind later */
int
ldap_unbind(Ldap *l)
{
	return request(l, Tunbind, seqno(l), nil, nil);
}

