/*
 * https://ldap.com/ldapv3-wire-protocol-reference-bind/
 */

#include <u.h>
#include <libc.h>
#include <auth.h>
#include <mp.h>
#include <ctype.h>
#include <fcall.h>
#include <thread.h>
#include <libsec.h>
#include <pool.h>
#include <9p.h>
#include <der.h>
#include "ldap.h"

struct Node {
	char *dn;			/* absolute distinguished name */
	void *buf;			/* file/dirent buffer */
	int len;			/* number of dirs/bytes in buf */
};

/* allocate a new node */
static Node *
mknode(char *dn)
{
	Node *n;

	n = emalloc9p(sizeof(Node));
	memset(n, 0, sizeof(Node));
	n->dn = estrdup9p(dn);
	return n;
}

static char *
newdn(char *dn, char *name, char *type)
{
	char *p;

	if(strcmp(name, "..") == 0){
		if((p = strchr(dn, ',')) == nil)
			return estrdup9p("");
		return estrdup9p(p+1);				/* +1 to step over the comma */
	}
	if((p = smprint("%s=%s,%s", type, name, dn)) == nil)
		sysfatal("No memory, %r");
	return p;
}

static void
responderrstr(Req *r)
{
	char e[ERRMAX];
	*e = 0;
	rerrstr(e, sizeof e);
	respond(r, e);
}

static int
dirgen(int slot, Dir *d, void *aux)
{
	Node *n;
	Dir *tab;

	n = aux;
	if(slot >= n->len)
		return -1;
	tab = n->buf;
	*d = tab[slot];
	d->uid = estrdup9p(d->uid);
	d->gid = estrdup9p(d->gid);
	d->muid = estrdup9p(d->muid);
	d->name = estrdup9p(d->name);
	return 0;
}

static void
fsattach(Req *r)
{
	Qid qid;
	Ldap *l;

	qid.path = PathRoot;
	qid.type = QTDIR;
	qid.vers = 1;

	l = r->srv->aux;
	r->ofcall.qid = qid;
	r->fid->qid = qid;
	r->fid->aux = mknode("");
	cache_incref(l->cache, "");
	respond(r, nil);
}

static char*
fsclone(Fid *ofid, Fid *fid)
{
	Node *on;
	Ldap *l;

	on = ofid->aux;
	l = fid->pool->srv->aux;

	fid->aux = mknode(on->dn);
	cache_incref(l->cache, on->dn);
	return nil;
}

static char*
fswalk1(Fid *fid, char *name, Qid *qp)
{
	Node *n;
	Ldap *l;
	char *dn;

	n = fid->aux;
	l = fid->pool->srv->aux;

	if(fid->qid.path == PathRoot)
		dn = srch_root_walk(l, n->dn, name, qp);
	else
		dn = srch_walk(l, n->dn, name, qp);
	if(dn == nil)
		return "not found";

	cache_decref(l->cache, n->dn);
	free(n->dn);	
	n->dn = dn;
	cache_incref(l->cache, n->dn);
	return nil;
}

static void
fsstat(Req *r)
{
	int rc;
	Ldap *l;
	Node *n;

	n = r->fid->aux;
	l = r->srv->aux;

	if(r->fid->qid.path == PathRoot){
		r->d.name = estrdup9p("");
		r->d.uid = estrdup9p("ldap");
		r->d.gid = estrdup9p("ldap");
		r->d.muid = estrdup9p("");
		r->d.mode = DMDIR | 0777;
		r->d.muid = estrdup9p("");
		r->d.mtime = r->d.atime = time(nil);
		r->d.qid.type = QTDIR;
		r->d.qid.path = PathRoot;
		r->d.qid.vers = 1;
		rc = 0;
	}
	else
	if(r->fid->qid.path == PathCtxt)
		rc = srch_root_stat(l, n->dn, &r->d);
	else
		rc = srch_stat(l, n->dn, &r->d);

	if(rc < 0)
		responderrstr(r);
	else
		respond(r, nil);
}

static void
fsopen(Req *r)
{
	Node *n;
	Ldap *l;
	
	l = r->srv->aux;
	n = r->fid->aux;

	if(r->fid->qid.type & QTDIR){
		if(r->fid->qid.path == PathRoot)
			n->len = srch_root_readdir(l, n->dn, &n->buf);
		else
			n->len = srch_readdir(l, n->dn, &n->buf);
	}
	else
		n->len = srch_read(l, n->dn, &n->buf);

	if(n->len < 0)
		responderrstr(r);
	else
		respond(r, nil);
}

static void
fsread(Req *r)
{
	Node *n;
	int xfer;
	uchar *buf;

	n = r->fid->aux;

	if(r->fid->qid.type & QTDIR){		/* dir read */
		dirread9p(r, dirgen, n);
		respond(r, nil);
		return;
	}

	if(r->ifcall.offset >= n->len){		/* EOF */
		r->ofcall.count = 0;
		respond(r, nil);
		return;
	}
	
	xfer = r->ifcall.count;
	if(r->ifcall.offset + r->ifcall.count >= n->len)
		xfer = n->len - r->ifcall.offset;

	buf = n->buf;
	memcpy(r->ofcall.data, buf + r->ifcall.offset, xfer);
	r->ofcall.count = xfer;
	respond(r, nil);
}

static void
fsdestroyfid(Fid *fid)
{
	Node *n;
	Ldap *l;

	n = fid->aux;
	l = fid->pool->srv->aux;
	if(n){
		cache_decref(l->cache, n->dn);
		free(n->buf);
		free(n->dn);
		free(n);
	}
}

static void
fsend(Srv *srv)
{
	Ldap *l;

	l = srv->aux;
	ldap_unbind(l);
	close(l->fd);
	free(l);
}

Srv fs = {
	.destroyfid = fsdestroyfid,
	.attach 	= fsattach,
	.open 		= fsopen,
	.read 		= fsread,
	.stat 		= fsstat,
	.clone 		= fsclone,
	.walk1 		= fswalk1,
	.end 		= fsend
};

Srv *fsp = &fs;
