#include <u.h>
#include <libc.h>
#include <auth.h>
#include <mp.h>
#include <libsec.h>
#include <der.h>
#include <fcall.h>
#include <thread.h>
#include <9p.h>
#include "ldap.h"

enum {
	QTBAD = 0xff			/* bad qid type, aka not filled in yet */
};

typedef struct Xdir Xdir;
typedef struct Xbuf Xbuf;

struct Xdir {
	Dir *dirtab;			/* array of Dirs allocated */
	int alloc;				/* Dirs allocated */
	int used;				/* Dirs in use (next free slot) */
	int current;			/* Dir currently being read */
};

struct Xbuf {				/* file/dir contents search */
	uchar *buf;				/* dir entries or file contents */
	int alloc;				/* # bytes / dir entries allocated */
	int used;				/* # bytes / dir entries in use */
};

static int isadir(Ldap *ldap, char *dn);

/* free allocated dir fields */
static void
freedir(Dir *d)
{
	free(d->name);
	free(d->uid);
	free(d->gid);
	free(d->muid);
}

/* fill in fields missing from ldap */
static void
fixupdir(Ldap *ldap, Dir *d)
{
	if(! d->atime)
		d->atime = time(nil);
	if(! d->mtime)
		d->mtime = d->atime;
	if(! d->uid)
		d->uid = estrdup9p("ldap");
	if(! d->gid)
		d->gid = estrdup9p(d->uid);

	if(d->qid.type == QTBAD){			/* no info yet */
		if(isadir(ldap, d->muid)){
			d->qid.type = QTDIR;
			d->mode = 0755 | DMDIR;
		}
		else{
			d->qid.type = QTFILE;
			d->mode = 0644;
		}
	}

	/* prevent clashes with our Magic paths - crude but good enough */
	if(d->qid.path == PathRoot || d->qid.path == PathCtxt)
		d->qid.path += 1024;

}

static Elem
filt_exist(void)
{
	return mkcontstr(QueryPresent, "objectclass");
}

static Elem
filt_match(char *name)
{
	return mkcontseq(QuerySubstring,
		mkel(mkoctstr("cn"),
			mkel(mkseq(
				mkel(mkcontstr(QueryAnd, name),nil)
			), nil)
		));
}

static Elem
attrs_stat(void)
{
	return mkseq(											/* attributes wanted */
		mkel(mkoctstr("vendorName"),						/* in root only */
		mkel(mkoctstr("vendorVersion"),						/* in root only */
		mkel(mkoctstr("namingContexts"),					/* in root only */
		mkel(mkoctstr("rootDomainNamingContext"),			/* in root only, in active directory only */
		mkel(mkoctstr("cn"),								/* file name */
		mkel(mkoctstr("ou"),								/* directory name */
		mkel(mkoctstr("name"),								/* misc name */
		mkel(mkoctstr("uSNChanged"),						/* serial no. of last change (file version) */
		mkel(mkoctstr("creatorsName"),						/* uid */
		mkel(mkoctstr("primaryGroupID"),					/* gid */
		mkel(mkoctstr("modifiersName"),						/* muid */
		mkel(mkoctstr("modifyTimestamp"),					/* mtime */
		mkel(mkoctstr("whenCreated"),						/* mtime, if no modifyTimestamp */
		mkel(mkoctstr("MsDS-Approx-Immed-Subordinates"),	/* # of children (is a dir) - approx, and AD only */
		nil)))))))))))))));
}

static Elem
attrs_walk(void)
{
	return mkseq(
		mkel(mkoctstr("uSNChanged"),						/* serial no. of last change (file version) */
		mkel(mkoctstr("MsDS-Approx-Immed-Subordinates"),	/* # of children (is a dir) - approx, and AD only */
	 	nil)));
}

static Elem
attrs_contexts(void)
{
	return mkseq(mkel(mkoctstr("namingContexts"), nil));
}

static Elem
attrs_all(void)
{
	return mkseq(mkel(mkoctstr("*"), nil));
}

/*************************************************************/

/* return the parent path (strip the first object from a DN path */
static char *
parent(char *dn)
{
	char *p;

	if((p = strchr(dn, ',')) == nil)
		sysfatal("'%s' contains no ','\n", dn);
	return p+1;
}

/* generate a hash of a string */
static uvlong
hash(Bytes *b)
{
	uchar digest[SHA1dlen];

	sha1(b->data, b->len, digest, nil);
	return *((uvlong *)digest) & ~0xfL;
}

/* get a large integer */
static vlong
intval(Bytes *b)
{
	int n;
	char str[64];

	n = sizeof(str);
	if(n > b->len)
		n = b->len;
	memcpy(str, b->data, n);
	str[n] = 0;
	return atoll(str);
}

/* extract a malloc'ed string from a value */
static char *
strval(Bytes *b)
{
	char *str;

	str = emalloc9p(b->len+1);
	memcpy(str, b->data, b->len);
	str[b->len] = 0;
	return str;
}

/*
 * extract the initial name, modifies the string passed in,
 * but returns a pointer to the same address (so it can be freed)
 */
static char *
dn2name(char *s)
{
	char *p;

	if((p = strchr(s, ',')) != nil)
		*p = 0;
	if((p = strchr(s, '=')) != nil)
		strcpy(s, p+1);
	return s;
}

/* true if name is the start of the DN */
static int
val_name(Namval *nv, char *s)
{
	char *p;

	if((p = strchr((char *)(nv->value->data), '=')) == nil)
		return 0;
	return memcmp(p+1, s, strlen(s)) == 0;
}

/* return true if ths param's name matches (case independent) */
static int
nam_is(Namval *nv, char *s)
{
	return cistrncmp(s, (char *)(nv->name->data), nv->name->len) == 0;
}

/* return true if ths param's value matches (case independent) */
static int
val_is(Namval *nv, char *s)
{
	return memcmp(s, nv->value->data, nv->value->len) == 0;
}

/*get another dir entry, growing the table as needed */
static Dir *
newdir(Xdir *xd)
{
	Dir *d;

	if(xd->used+1 >= xd->alloc){
		xd->alloc += 8;
		xd->dirtab = erealloc9p(xd->dirtab, xd->alloc * sizeof(Dir));
	}
	xd->current = xd->used++;

	d = &xd->dirtab[xd->current];
	memset(d, 0, sizeof(Dir));
	d->qid.type = QTBAD;
	return d;
}

/*****************************
 * handle walk/stat/readdir requests at the root level.
 * root is special in ldap. 
 */

static int
root_attrib(void *aux, Namval *nv)
{
	Dir *d;

	if(nam_is(nv, "namingContexts")){
		d = newdir(aux);
		d->qid.type = QTDIR;
		d->qid.path = PathCtxt;
		d->qid.vers = 1;
		d->muid = strval(nv->value);
		d->mode = 0755 | DMDIR;
		d->name = dn2name(strval(nv->value));
	}
	return 0;
}

static int
root_readdir(Ldap *ldap, char *dn)
{
	int i;
	Xdir xdir;
	Srch srch;

	memset(&xdir, 0, sizeof(xdir));
	memset(&srch, 0, sizeof(srch));

	srch.aux = &xdir;
	srch.attrib = root_attrib;
	if(ldap_search(ldap, "", attrs_contexts(), ScopeBase, filt_exist(), 0, &srch) < 0){
		for(i = 0; i < xdir.used; i++)
			freedir(&xdir.dirtab[i]);
		return -1;
	}

	for(i = 0; i < xdir.used; i++)
		fixupdir(ldap, &xdir.dirtab[i]);

	ldap->cache = cache_fill(ldap->cache, dn, xdir.used, xdir.dirtab);
	return 0;
}

int
srch_root_readdir(Ldap *ldap, char *dn, Dir **dirs)
{
	int num;

	if(Verbose)
		fprint(2, "\nsrch_root_readdir %s\n", dn);

	if((num = cached_readdir(ldap->cache, dn, dirs)) >= 0)
		return num;
	if(root_readdir(ldap, dn) < 0)
		return -1;
	return cached_readdir(ldap->cache, dn, dirs);
}

static void
gen_root(Ldap *ldap)		/* root dir is special, hand craft it */
{
	Dir dir;

	memset(&dir, 0, sizeof(dir));
	dir.qid.type = QTDIR;
	dir.qid.path = PathCtxt;
	dir.qid.vers = 1;
	dir.muid = estrdup9p("");
	dir.mode = 0755 | DMDIR;
	dir.name = estrdup9p("");
	fixupdir(ldap, &dir);
	ldap->cache = cache_fill(ldap->cache, "", 1, &dir);
}

int
srch_root_stat(Ldap *ldap, char *dn, Dir *d)
{

	if(Verbose)
		fprint(2, "\nsrch_root_stat %s\n", dn);

	if(cached_stat(ldap->cache, dn, d) == 0)
		return 0;
	if(strcmp(dn, "") == 0)
		gen_root(ldap);
	else
	if(root_readdir(ldap, dn) < 0)
		return -1;
	return cached_stat(ldap->cache, dn, d);
}

char *
srch_root_walk(Ldap *ldap, char *dn, char *name, Qid *q)
{
	char *ndn;

	if(Verbose)
		fprint(2, "\nsrch_root_walk %s\n", dn);

	if((ndn = cached_walk(ldap->cache, dn, name, q)) != nil)
		return ndn;
	if(root_readdir(ldap, dn) < 0)
		return nil;
	return cached_walk(ldap->cache, dn, name, q);
}

/*****************************
 * detect directories by searching for any objects under the current level
 */
static int
isadir_object(void *aux, Bytes *)
{
	int *h;

	h = aux;
	(*h)++;
	return 0;
}

static int
isadir(Ldap *ldap, char *dn)
{
	int hits;
	Srch srch;

	if(Verbose)
		fprint(2, "\nsrch_isadir dn=%s\n", dn);

	hits = 0;
	memset(&srch, 0, sizeof(srch));

	srch.aux = &hits;
	srch.object = isadir_object;
	ldap_search(ldap, dn, attrs_walk(), ScopeSingle, filt_exist(), 2, &srch);

	if(Verbose)
		fprint(2, "isadir: dn=%s hits=%d\n", dn, hits);

	return hits;
}

static int
other_object(void *aux, Bytes *dn)
{
	Dir *d;

	d = newdir(aux);
	d->qid.path = hash(dn);
	d->name = dn2name(strval(dn));
	d->muid = strval(dn);
	return 0;
}

static int
other_attrib(void *aux, Namval *nv)
{
	Dir *d;
	Xdir *xd;

	xd = aux;
	d = &(xd->dirtab[xd->current]);

	if(nam_is(nv, "MsDS-Approx-Immed-Subordinates")){
		if(intval(nv->value) == 0){
			d->qid.type = QTFILE;
			d->mode = 0644;
		}
		else{
			d->qid.type = QTDIR;
			d->mode = 0755 | DMDIR;
		}
	}

	if(nam_is(nv, "namingContexts")){				/* only appears at root level */
		free(d->name);
		d->name = dn2name(strval(nv->value));
	}
	if(! d->qid.vers && nam_is(nv, "uSNChanged"))
		d->qid.vers = intval(nv->value);
	if(! d->uid && nam_is(nv, "creatorsName"))
		d->uid = dn2name(smprint("%L", nv));
	if(! d->gid && nam_is(nv, "primaryGroupID"))
		d->gid = dn2name(smprint("%L", nv));
	if(! d->mtime && nam_is(nv, "modifyTimestamp"))
		d->mtime = iso2epoch(nv->value);
	return 0;
}


static int
other_readdir(Ldap *ldap, char *dn)
{
	int i;
	Xdir xdir;
	Srch srch;

	memset(&xdir, 0, sizeof(xdir));
	memset(&srch, 0, sizeof(srch));

	srch.aux = &xdir;
	srch.object = other_object;
	srch.attrib = other_attrib;
	if(ldap_search(ldap, dn, attrs_stat(), ScopeSingle, filt_exist(), 0, &srch) < 0){
		for(i = 0; i < xdir.used; i++)
			freedir(&xdir.dirtab[i]);
		return -1;
	}

	for(i = 0; i < xdir.used; i++)
		fixupdir(ldap, &xdir.dirtab[i]);

	ldap->cache = cache_fill(ldap->cache, dn, xdir.used, xdir.dirtab);
	return 0;
}

int
srch_readdir(Ldap *ldap, char *dn, Dir **dirs)
{
	int num;

	if(Verbose)
		fprint(2, "\nsrch_readdir %s\n", dn);

	if((num = cached_readdir(ldap->cache, dn, dirs)) >= 0)
		return num;
	if(other_readdir(ldap, dn) < 0)
		return -1;
	return cached_readdir(ldap->cache, dn, dirs);
}

int
srch_stat(Ldap *ldap, char *dn, Dir *d)
{
	if(Verbose)
		fprint(2, "\nsrch_stat %s\n", dn);

	if(cached_stat(ldap->cache, dn, d) == 0)
		return 0;
	if(other_readdir(ldap, dn) < 0)
		return -1;
	return cached_stat(ldap->cache, dn, d);
}

char *
srch_walk(Ldap *ldap, char *dn, char *name, Qid *q)
{
	char *ndn;

	if(Verbose)
		fprint(2, "\nsrch_walk %s\n", dn);

	if((ndn = cached_walk(ldap->cache, dn, name, q)) != nil)
		return ndn;
	if(other_readdir(ldap, dn) < 0)
		return nil;
	return cached_walk(ldap->cache, dn, name, q);
}


/*****************************
 * read a file
 */
static int
read_attrib(void *aux, Namval *nv)
{
	Fmt *f;

	f = aux;
	fmtprint(f, "%B=%L\n", nv->name, nv);
	return 0;
}

int
srch_read(Ldap *ldap, char *dn, char **buf)
{
	Fmt fmt;
	Srch srch;

	if(Verbose)
		fprint(2, "\nsrch_read %s\n", dn);

	fmtstrinit(&fmt);
	memset(&srch, 0, sizeof(srch));

	srch.aux = &fmt;
	srch.object = nil;
	srch.attrib = read_attrib;
	if(ldap_search(ldap, dn, attrs_all(), ScopeGlobal, filt_exist(), 0, &srch) < 0){
		free(fmtstrflush(&fmt));
		return -1;
	}

	*buf = fmtstrflush(&fmt);
	return strlen(*buf);
}

/*****************************
 * search for a object
 * FIXME: Not fully implemented!
 */
int
srch_look(Ldap *ldap, char *dn, char *name)
{
	Srch srch;
	memset(&srch, 0, sizeof(srch));

	return ldap_search(ldap, dn, attrs_stat(), ScopeGlobal, filt_match(name), 0, &srch);
}


/*****************************
 * system info
 */
static int
system_attrib(void *, Namval *)
{
	return 0;
}

int
srch_config(Ldap *ldap)
{
	Srch srch;

	memset(&srch, 0, sizeof(srch));

	srch.attrib = system_attrib;
	return ldap_search(ldap, "", attrs_all(), ScopeBase, mkcontstr(QueryPresent, ""), 0, &srch);
}

